/*
    pkdump.c - Port scanning Detection version 3.3
    By  Fabio Salvucci (dsmoker at users.sourceforge.net) 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include <sys/socket.h>
#include <features.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <linux/if.h>

#include <pthread.h>

#include <signal.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>


#define ETHMINLEN    46
#define IPV4HEADER   20
#define IPV6HEADER   40
#define IPV6FRAG     8
#define PCKT         65536 

struct HEADER {  
    
        unsigned char *packet_data ;
   
        unsigned int  packet_length;
   
        struct OVERWRITE *overwrite;

        unsigned int  frag_number;
   
        struct HEADER *next ;

}*Start = NULL;

pthread_mutex_t mutex  = PTHREAD_MUTEX_INITIALIZER ;

pthread_cond_t writing = PTHREAD_COND_INITIALIZER ;

pthread_cond_t reading = PTHREAD_COND_INITIALIZER ;

int READWRITE = 0 ;

char *IFACE ;
struct in_addr  Iddr;
struct in_addr  *iddr  = &Iddr;
struct in6_addr Iddr6;
struct in6_addr *iddr6 = &Iddr6;
 
int IGNORE_FLAG = 1 ;
int OUTGOING_FLAG = 0 ;
unsigned short int L_FLAGS = 0 ;


struct OVERWRITE {
   
       unsigned char      *packet_data ;

       unsigned short int offset ;
   
       unsigned int       length ;

       struct OVERWRITE   *next;
};  


struct FD {
   
       int   fd ;
       char  interface_name[20];  
}Fd;

pthread_t thread1ID ,thread2ID ;   


char *Flags_S_name[] = { ":FIN" ,":SYN" ,":RST" ,":PSH" ,
                         ":ACK" ,":URG" ,NULL     };

unsigned short int Flags_S_value[] = { 0x1 ,0x2 ,0x4 ,0x8 ,0x10 ,0x20 ,0x0 }; 
    

char *Flags_name[] = {"DUMMY","NULL" ,"UDPNULL" ,"UDPPORT" ,
                      "FIN","SYN","RST","PSH","ACK","URG","XMAS" ,"MYMO" ,
                      NULL 
                     }; 


unsigned short int Flags_value[] ={ 0xFF ,0x40 ,0x41 ,0x42 ,
                                    0x1 ,0x2 ,0x4 ,0x8 ,0x10 ,0x20, 0x29 ,0x14,
                                    0x0
                                  };
enum Options_alias { INTCP = 1 ,INUDP ,OUTTCP ,OUTUDP ,ALLTCP ,ALLUDP };   
char *Options_name[] = { "DUMMY" ,"-iT" ,"-iU" ,"-oT","-oU" ,
                                  "-aT" ,"-aU" ,NULL };
unsigned short int  Options_value[] = { 0xFF ,0x01 ,0x02 ,0x03 ,0x04,
                                              0x05 ,0x06 , 0x00 };  
   

enum IPversion_alias { OIPV4 = 1 ,OIPV6 };
char *IPversion_name [] = { "DUMMY" ,"IPV4" ,"IPV6" ,NULL };
unsigned short int IPversion_value [] = { 0xFF ,0x01 ,0x02 ,0x00 };   
   
enum Direction_alias { INGOING=0 ,OUTGOING };
char *Direction_name[] = { "INGOING" ,"OUTGOING" };  
   


struct FLAGS {
   
       unsigned short int  flags;
       struct FLAGS       *next ;

}*Flags=NULL;

struct PORT {
   
       unsigned int          port ;
       struct PORT           *next ;
};

struct OPTIONS {
       
       unsigned short int     opt ; 
       unsigned       int     sport ;
       unsigned       int     eport ;
       struct   PORT          *rport ;
       unsigned int           myaddress ;
       unsigned short int     myaddress6[8];
       unsigned short int     mask ;
       struct                 OPTIONS *next;

}*Options=NULL ;

struct IPV {
   
       unsigned short int ver ;
}Ipv;


char *Directory = NULL ;


FILE *Out1 = NULL ;


char *stcat ( char *p ,char *q ,int slash ) {
   
     char *data ,*dummy ,*point1 = p ,*point2 = q;
   
     if ( !q || !p ) { printf ("\nERROR :No directory \n\n");exit(1);}
   
     if ((data = (char*)malloc( strlen(p) + strlen (q) + 4 )) == NULL ){
	   
	       fprintf (stdout ,"\nERROR :No memory \n\n") ;
	      
	       exit(1);
     }
   	
     dummy = data ;
	
     while ( *point1 ) { 
	
	if ( iscntrl(*point1) ) {point1++;continue ;} 
	
	*dummy = *point1 ;dummy++ ;
	
	point1++;
     }
   
     if ( slash ) *(dummy++) = '/' ;
     
     while ( *point2 ) { 
	
	if ( iscntrl (*point2) ) {point2++;continue ;}
	
	*dummy = *point2 ;dummy++ ;
	     
        point2++;
      
     }
     
     *(dummy++) = '\0' ;
   
     return (data);
   
}     
	   
   

int if_pr_set (char *if_interface_name) 
{

 int fd = 0 ;
 static struct ifreq ifr ;
 int s ;
    
	  
 fd = socket ( PF_PACKET ,SOCK_DGRAM , htons (ETH_P_ALL) );  

 fprintf ( stdout ,
    "\n\nALERT : Make sure that the Interface carry an IP,IPV6 datagram\n\n"); 
   
   
 if (fd < 0 ) {
     
         fprintf (stdout ,"\nERROR :Can't get PACKET SOCKET socket\n"); 
         fprintf (stdout ,"compile packet socket in your kernel or\n ");
         fprintf (stdout ,"insert it as a module ( af_packet.o ) \n\n" );
         exit(0);
  }

 strcpy (ifr.ifr_name ,if_interface_name );
 strcpy (Fd.interface_name ,if_interface_name );
   
 s = ioctl ( fd ,SIOCGIFFLAGS ,&ifr );
   
 if ( s<0 ) {  
    close (fd); 
    fprintf (stdout ,"\nERROR :Can't find interface \n\n");
    exit (0);
 }

 if ( !(ifr.ifr_flags & IFF_UP) ) {
    close (fd) ;
    fprintf (stdout ,"\nERROR :Interface is DOWN \n\n");
    exit(0);
 }
   
 ifr.ifr_flags |= (IFF_PROMISC|IFF_DYNAMIC|IFF_DEBUG) ;
   
 s = ioctl ( fd ,SIOCSIFFLAGS ,&ifr ) ;
   
 if ( s < 0 ) {
    close(fd) ;
    fprintf (stdout , "\nERROR :Can't set interface flags \n\n" );
    exit(0);
 }
   
    
 s = ioctl ( fd ,SIOCGIFADDR ,&ifr ) ; 
   
 if ( s < 0 ) {
    fprintf (stdout , "\nERROR :Can't find interface address\n\n" );
    exit(0);
 }  
 
 if (ifr.ifr_addr.sa_family==AF_INET) 
        memcpy ((void*)&(iddr->s_addr),(void*)&(ifr.ifr_addr.sa_data[2]) ,4 );
  
 return (fd) ;  
   
}

char *reverse ( char *s ) 
{
  int c ,i ,j ;

  for ( i = 0 ,j = strlen (s) - 1 ;i<j ;i++ ,j-- ) {
      
      c = *(s+i);

      *(s+i) = *(s+j) ;

      *(s+j) = c;
  }
  return (s);
}  
 
char *citoa ( unsigned char  n ,char *s ) 
{
  unsigned char  i ,dummy ;

  i =0 ;

  do {

     dummy = n % 16 + '0' ;
     
     if ( dummy >= 0x3A ) dummy += 7 ;
     
     *(s+(i++)) = dummy ;

     if ( n == 0 ) break ;
     
  } while ( ( n /= 16 ) > 0 );

  *(s+i) = '\0' ;

  if ( strlen (s) == 1 ) { *(s+1) = '0' ;*(s+2) = '\0' ;} 
   
  s = reverse ( s ) ;

  return ( s ) ;
}   

char *iitoa ( unsigned int n ,char *s ) 
{
  unsigned char  i ,dummy ;

  i =0 ;

  do {

     dummy = n % 16 + '0' ;
     
     if ( dummy >= 0x3A ) dummy += 7 ;
     
     *(s+(i++)) = dummy ;
     
     if ( n == 0 ) break ;
     
  } while ( ( n /= 16 ) > 0 );

  *(s+i) = '\0' ;

  if ( strlen (s) == 1 ) { *(s+1) = '0' ;*(s+2) = '\0' ;} 
   
  s = reverse ( s ) ;

  return ( s ) ;
}   


char *ip_disasm ( char *frag ,char *where)
{

   unsigned char dummy ;
   char *ffrag = frag  ;
   char *value = where ;
     
     
   while ( *ffrag ) {
   
         if ( *ffrag >= 'A' ) dummy = *ffrag - 55 ;	      
         else dummy = *ffrag - 48 ;
	      	      
         dummy *= 16 ;
	      
	      
         if ( *(ffrag+1) >= 'A' ) dummy += *(ffrag+1) - 55 ;
         else dummy += *(ffrag+1) - 48 ;
	      
	 *(value++) = dummy ;     

         ffrag += 2;
   }

   return (where);

}



struct HEADER *ip_assembly ( unsigned char *packet_fragment ,int flush )
{
   struct OVERWRITE *sover ,*over ;   
   struct HEADER    *head ;

   struct Ufrag {

          unsigned int       length ;
          unsigned short int offset ;
          unsigned char     *packet ;
          struct Ufrag      *next   ;
   };
   
   struct Mfrag {
          struct tm          *loctime ;
          unsigned int       total_length ;
          unsigned int       identification ;
          unsigned short int header ;
          struct Ufrag       *ufrag ;
          struct Mfrag       *next  ;
          struct Mfrag       *prev  ;
   };	
      
   static struct Mfrag *mfrag = NULL ;
   struct Mfrag *mpck = NULL ,*mummy = NULL ,*mummy0 = NULL ;
   struct Ufrag *upck = NULL ,*uummy1 = NULL ,*uummy2 = NULL ;  

   char *pkbuild = NULL ;
      
   unsigned short int header = 0 ,offset = 0 ,mfragment = 0 ,version = 0;
   
   unsigned int length = 0 ,id = 0 ,tlen = 0 ,frag_number = 0 ;
   
   char   byte[3] ,*tmp;
   int    count = 0 ,mcount = 0 ,ccount = 0 ,maxccount = 0 ;
   
   unsigned short int ident ,tl ,fo ;
   unsigned char next_header ;

   unsigned int minpos  = 0 ;
   unsigned int maxpos  = 0 ;
   unsigned int minover = 0 ;
   unsigned int maxover = 0 ;
   unsigned int mmflag  = 0 ;
   unsigned char *over_pkt = NULL ; 
   
   struct tm *loctime ;
   time_t curtime ;
         
   curtime = time (NULL);  
   
   loctime = localtime (&curtime);    

   if ( flush ) goto flush0 ;
   
   version = packet_fragment[0] >> 4 ; 

   if ( version == 4 ) {

      header = ( packet_fragment[0] & 0x0F  ) * 4  ; 
   
      length   = packet_fragment[2] ;	      
      length <<= 8 ;	
      length  += packet_fragment[3] ;
      
      id   = packet_fragment[4];
      id <<=8;
      id  += packet_fragment[5];   
      
      offset   = packet_fragment[6] & 0x1F ;
      offset <<= 8 ;
      offset  += packet_fragment[7] ;
      
      mfragment = packet_fragment[6] & 0x20  ;
      
   }
   else if ( version == 6 ) {
      
        header = IPV6HEADER ;
      
        length   = packet_fragment[4] ;	      
        length <<= 8 ;	
        length  += packet_fragment[5] ;
        length  += IPV6HEADER ;

        id   = packet_fragment [header + 4] ;
	id <<= 8 ;
	id  += packet_fragment [header + 5] ;
	id <<= 8 ;
	id  += packet_fragment [header + 6] ;
	id <<= 8 ;
	id  += packet_fragment [header + 7] ;

        offset = 0 ;
      
        if ( packet_fragment[6] == IPPROTO_FRAGMENT ){
      
	   offset   = packet_fragment[header + 2] ;
	   offset <<= 8 ;
	   offset  += packet_fragment[header + 3] ;
	   offset >>= 3;
	   
	   mfragment = packet_fragment [header + 3] & 0x01 ;
          
	   next_header = packet_fragment[header] ;
	   
	   memcpy ( &packet_fragment[header] ,
		    &packet_fragment[header+8] ,
		    length - IPV6HEADER + 8 ) ;
	   
	   length -= 8 ;
	  
       }
       
       
   }	
   
   if (!mfrag) {

      if ( (mfrag=mpck=(struct Mfrag*) malloc ( sizeof (struct Mfrag)))==NULL){
	 
	 fprintf (stdout , "\nERROR :No space for fragment 1\n\n" );
	 
	 exit(1);
      }

      if ( (mpck->loctime = (struct tm*) malloc (sizeof(struct tm)))==NULL ){
	 
	 fprintf (stdout , "\nERROR :No space for time structure 1\n\n");
	 
	 exit(1);
      }	   
      
      memcpy ((void*)mpck->loctime ,(void*) loctime ,sizeof (struct tm));

      if (!mfragment) {
	 if (!offset) mpck->total_length = length - header ;
         else mpck->total_length = offset + length - header ;
      }
      else mpck->total_length = 0 ;
      
      mpck->identification   = id;
      
      if (!offset) mpck->header = header;
      else mpck->header = 0 ;

      if ( (mpck->ufrag=(struct Ufrag*) malloc ( sizeof(struct Ufrag)))==NULL){
	 
	 fprintf ( stdout ,"\nERROR :No space for fragment 2\n\n" );
	 
	 exit(1);
      }

      if (!offset) {
	   
         mpck->ufrag->length = length ;
         mpck->ufrag->offset = offset ;
      
         if ( (mpck->ufrag->packet=(char*)malloc (length+1)) == NULL){
	 
	    fprintf ( stdout ,"\nERROR :No space for fragment 3\n\n" );
	 
	    exit(1);
         }

	 if (version == 6 && mfragment) packet_fragment[6] = next_header ;
	 
         memcpy ( mpck->ufrag->packet ,packet_fragment ,length );
      }
      else {
	   
	   mpck->ufrag->length = length - header  ;
           mpck->ufrag->offset = offset ;
      
           if ( (mpck->ufrag->packet=(char*)malloc (length-header+1)) == NULL){
	 
	      fprintf ( stdout ,"\nERROR :No space for fragment 4\n\n" );
	 
	      exit(1);
           }

           memcpy ( mpck->ufrag->packet ,&packet_fragment[header] ,
		                                             length - header );
      }
      
      mpck->ufrag->next = NULL ;
      
      mpck->next = NULL ;
      mpck->prev = NULL ;
						 
   }
   else { /*mfrag*/

      mpck = mfrag ;
      
      while (mpck) {

	    if ( mpck->identification == id ) break ;
	 
	    mpck = mpck->next ;
      }
      
      if ( mpck ) {
	 
       if (!mfragment) {
	  if (!offset) mpck->total_length = length - header ;
          else mpck->total_length = offset + length - header ;
       }
             	 	 
       if (!offset) mpck->header = header;	 
	 
       upck = mpck->ufrag ;
	 
       while ( upck->next ) upck = upck->next ;
	 
       if ( (upck->next=(struct Ufrag*) malloc ( sizeof(struct Ufrag)))==NULL){
	 
	 fprintf  (stdout , "\nERROR :No space for fragment 5\n\n" );
	 
	 exit(1);
       }

       upck = upck->next ;
	 
       if (!offset) {
	  
             upck->length = length ;
             upck->offset = offset ;
      
             if ( (upck->packet=(char*)malloc (length+1)) == NULL){
	 
	        fprintf (stdout ,"\nERROR :No space for fragment 6\n\n" );
	 
	        exit(1);
             }
	     
	     if(version == 6 && mfragment) packet_fragment[6]=next_header ;
           	     
	     memcpy ( upck->packet ,packet_fragment ,length );
	       
       }
       else {	 
	      
             upck->length = length - header ;
             upck->offset = offset ;
      
             if ( (upck->packet=(char*)malloc (length-header+1)) == NULL){
	 
	        fprintf (stdout ,"\nERROR :No space for fragment 7\n\n" );
	 
	        exit(1);
             }
	     
	     memcpy ( upck->packet ,&packet_fragment[header] ,length-header );
       }
	 
       upck->next = NULL ;
      	 
      }
      else { /*mpck*/

       mpck = mfrag;
	 
       while ( mpck->next ) mpck = mpck->next ;	 
	 
       if ((mpck->next=(struct Mfrag*) malloc ( sizeof(struct Mfrag)))==NULL){
	 
	  fprintf (stdout , "\nERROR :No space for fragment 8\n\n" );
	 
	  exit(1);
       
       }	 	 

       mummy = mpck ;	 
	 
       mpck = mpck->next ;
	 
       mpck->prev = mummy ;
	 
       mpck->next = NULL ;

       if ( (mpck->loctime = (struct tm*) malloc (sizeof(struct tm)))==NULL){
	 
	 fprintf (stdout , "\nERROR :No space for time structure 1\n\n");
	 
	 exit(1);
       }	   
      
       memcpy ((void*)mpck->loctime ,(void*) loctime ,sizeof (struct tm));

       if (!mfragment) {
	 if (!offset) mpck->total_length = length - header ;
         else mpck->total_length = offset + length - header ;
       }
       else mpck->total_length = 0 ;	 
	 
       mpck->identification   = id;	 

       if (!offset) mpck->header = header;
       else mpck->header = 0 ;

       if ((mpck->ufrag=(struct Ufrag*) malloc ( sizeof(struct Ufrag)))==NULL){
	 
	 fprintf ( stdout ,"\nERROR :No space for fragment 9\n\n" );
	 
	 exit(1);
       }

       if (!offset) {
	   
         mpck->ufrag->length = length ;
         mpck->ufrag->offset = offset ;
      
         if ( (mpck->ufrag->packet=(char*)malloc (length+1)) == NULL){
	 
	    fprintf ( stdout ,"\nERROR :No space for fragment 10\n\n" );
	 
	    exit(1);
         }

         if (version == 6 && mfragment ) packet_fragment[6] = next_header ;
       
         memcpy ( mpck->ufrag->packet ,packet_fragment ,length );
       }
       else {
	   
	   mpck->ufrag->length = length - header  ;
           mpck->ufrag->offset = offset ;
      
           if ( (mpck->ufrag->packet=(char*)malloc (length-header+1)) == NULL){
	 
	      fprintf ( stdout ,"\nERROR :No space for fragment 11\n\n" );
	 
	      exit(1);
           }

           memcpy ( mpck->ufrag->packet ,&packet_fragment[header] ,
		                                             length - header );
       }
	 
       mpck->ufrag->next = NULL ;
      
      }
      
      
   }

flush0:   
   
   mpck = mfrag ;
   
   while (mpck) {

     minpos = maxpos = mmflag = maxccount = 0 ; 

     sover = over = NULL ; 
      
     uummy1 = mpck->ufrag ;
	 
     tlen = 0 ;	

     while ( uummy1 ) {tlen += uummy1->length ;uummy1 = uummy1->next ;}
      
     uummy1 = mpck->ufrag ; 
	
     if ( tlen == mpck->total_length + mpck->header ) {

go:    if (version == 4 ) {
	  
	    if ( (pkbuild=(char*)malloc ((PCKT*2)+(mpck->header*2)+1 )) == 0){
	 
	    fprintf ( stdout ,"\nERROR :No space for fragment 12\n\n" );
	 
	    exit(1);
	    }
            memset ( pkbuild , ' ' , sizeof (pkbuild) );
            *(pkbuild+(PCKT*2)+(mpck->header*2)) = '\0';
       }
       else {
	    
	    if ( (pkbuild=(char*)malloc ((PCKT*2)+(IPV6HEADER*2)+1 )) == 0){
	 
	    fprintf ( stdout ,"\nERROR :No space for fragment 13\n\n" );
	 
	    exit(1);
	    }
	    memset ( pkbuild , ' ' , sizeof (pkbuild) );
	    *(pkbuild+(PCKT*2)+(IPV6HEADER*2)) = '\0';
       }
   
       frag_number = 0 ;
	
       while (uummy1) {
	 
	if (version == 4) { 
	       
            if ( (uummy1->offset + uummy1->length ) >= PCKT+mpck->header ) {  
              fprintf (stdout, "\nALERT :Datagram out of the size !!\n\n") ;
	      fprintf (stdout,"\a\a");
	      fflush  ( stdout );
	      *(pkbuild+(PCKT*2)+(mpck->header*2)) = '\0';
	      break ;
            }
	}
	else { 
	       
            if ( (uummy1->offset + uummy1->length ) >= PCKT+IPV6HEADER ) {  
              fprintf (stdout, "\nALERT :Datagram out of the size !!\n\n") ;
	      fprintf (stdout,"\a\a");
	      fflush  ( stdout );
	      *(pkbuild+(PCKT*2)+(IPV6HEADER*2)) = '\0';
	      break ;
            }
	}
	    
	if (mmflag) {
	    
	 if ( uummy1->offset != 0 ) {
	    
	    minover = (mpck->header + uummy1->offset ) * 2 ;
	    maxover = (mpck->header + uummy1->offset + uummy1->length )*2;
	 }
	 else {   
	      minover = 0;
	      maxover = (uummy1->length ) * 2 ;	
	 }

	 if ( minover == minpos) {	
	
	    if (!sover) { 
	       
               if ((sover=over =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite1 \n\n" );
	 
	          exit(1);
	       }
	       over->next = NULL ;
		    
	     }
	     else {	   
		      
	       if ((over->next =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite2 \n\n" );
	 
	          exit(1);
	       }
	       over = over->next ;
	       over->next = NULL ;
	       	
	     }
		     
	     if ((over->packet_data =(char*)malloc((maxover-minpos)/2))==0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite3 \n\n" );
	 
	          exit(1);
	     }
	     if ((over_pkt =(char*)malloc(maxover-minpos+1)) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite4 \n\n" );
	 
	          exit(1);
	     }	
	     memset ( over_pkt ,'\0' , maxover-minpos+1 );
	     memcpy ( over_pkt ,pkbuild+minpos ,maxover-minpos);
		
	     over->packet_data = ip_disasm ( over_pkt ,over->packet_data );
	
	     free((void*)over_pkt);	
		
	     if (minpos) over->offset = minpos / 2 ;
	     else over->offset = 0 ;
	    
	     over->length = (maxover-minpos)/2 ;
	    
	 }   
	 else if ( (minover > minpos) && (maxover < maxpos) ) { 
	
	    if (!sover) { 
	       
               if ((sover=over =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite5 \n\n" );
	 
	          exit(1);
	       }
	       over->next = NULL ;
		    
	     }
	     else {	   
		      
	       if ((over->next =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite6 \n\n" );
	 
	          exit(1);
	       }
	       over = over->next ;
	       over->next = NULL ;
	       	
	     }
		     
	     if ((over->packet_data =(char*)malloc((maxover-minover)/2))==0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite7 \n\n" );
	 
	          exit(1);
	     }
	     if ((over_pkt =(char*)malloc(maxover-minover+1)) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite8 \n\n" );
	 
	          exit(1);
	     }	
	     memset ( over_pkt ,'\0' , maxover-minover+1 );
	     memcpy ( over_pkt ,pkbuild+minover ,maxover-minover);
		
	     over->packet_data = ip_disasm ( over_pkt ,over->packet_data );
	
	     free((void*)over_pkt);	
		
	     if (minover) over->offset = (minover/2)+1 ;
	     else over->offset = 0 ;
	    
	     over->length = (maxover-minover)/2 ;
		
	 } 
	 else if ( (minover < minpos) && (maxover > minpos) ) { 
	
	    if (!sover) { 
	       
               if ((sover=over =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite9 \n\n" );
	 
	          exit(1);
	       }
	       over->next = NULL ;
		    
	     }
	     else {	   
		      
	       if ((over->next =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite10 \n\n" );
	 
	          exit(1);
	       }
	       over = over->next ;
	       over->next = NULL ;
	       	
	     }
		     
	     if ((over->packet_data =(char*)malloc((maxover-minpos)/2))==0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite11 \n\n" );
	 
	          exit(1);
	     }
	     if ((over_pkt =(char*)malloc(maxover-minpos+1)) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite12 \n\n" );
	 
	          exit(1);
	     }	
	     memset ( over_pkt ,'\0' , maxover-minpos+1 );
	     memcpy ( over_pkt ,pkbuild+minpos ,maxover-minpos);
		
	     over->packet_data = ip_disasm ( over_pkt ,over->packet_data );
	
	     free((void*)over_pkt);	
		
	     if (minpos) over->offset = minpos / 2 ;
	     else over->offset = 0 ;
	    
	     over->length = (maxover-minpos)/2 ;
		
	 } 
	 else if ( (minover < maxpos) && (maxover > maxpos) ) { 
	
	    if (!sover) { 
	       
               if ((sover=over =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite13 \n\n" );
	 
	          exit(1);
	       }
	       over->next = NULL ;
		    
	     }
	     else {	   
		      
	       if ((over->next =
		  (struct OVERWRITE*)malloc(sizeof(struct OVERWRITE))) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite14 \n\n" );
	 
	          exit(1);
	       }
	       over = over->next ;
	       over->next = NULL ;
	       	
	     }
		
	     if ((over->packet_data =(char*)malloc((maxpos-minover)/2))==0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite15 \n\n" );
	 
	          exit(1);
	     }
	     if ((over_pkt =(char*)malloc(maxpos-minover+1)) == 0){
	 
	          fprintf ( stdout ,"\nERROR :No space for overwrite16 \n\n" );
	 
	          exit(1);
	     }	
	     memset ( over_pkt ,'\0' , maxpos-minover+1 );
	     memcpy ( over_pkt ,pkbuild+minover ,maxpos-minover);
		
	     over->packet_data = ip_disasm ( over_pkt ,over->packet_data );
	 
             free((void*)over_pkt);
		
	     if ( minover ) over->offset = (minover/2)+1 ;
	     else over->offset = 0 ;
	    
	     over->length = (maxpos-minover)/2 ;
             
	 }
	   
	}
	  	  
	if (uummy1->offset != 0) ccount = (mpck->header+uummy1->offset)*2;
	else ccount = 0;  
		     	  
	if ( ccount < minpos ) minpos = ccount ;   
	  
        for ( mcount = 0 ;
	      mcount < uummy1->length  ;
	      mcount ++ ,ccount += 2         ) {
	   
	      tmp = citoa ( uummy1->packet[mcount] ,byte );
	 
	      *(pkbuild+ccount)   = *tmp;
	                  
              *(pkbuild+ccount+1) = *(tmp+1) ;
	  
         }

	  
	 if ( ccount > maxccount) maxccount = ccount + 2  ; 

         if ( ccount > maxpos ) maxpos = ccount ;	 	   

	 mmflag = 1 ; 

	 frag_number++; 
	  
	 uummy1 = uummy1->next;
	  
       }	

       *(pkbuild+maxccount) = '\0' ;	
	
       if ((head = (struct HEADER*)malloc (sizeof (struct HEADER))) == NULL){  

	    fprintf ( stdout ,"\nERROR :No space for header \n\n" );
	 
	    exit(1);
       }
	  
       head->packet_length = strlen (pkbuild) / 2 ;
	 
       if (( head->packet_data = 
	                (char*) malloc (head->packet_length+1)) == NULL ) {
              fprintf (stdout, "\nERROR :No space for fragment 14\n\n") ;
              exit (1) ;
       }
	 
       head->frag_number = frag_number ;
	 
       head->packet_data = ip_disasm ( pkbuild ,head->packet_data);
	 	 
       if (sover) head->overwrite = sover ;
       else head->overwrite = NULL ;
	
       free ((void*) pkbuild);	
	
       if ( mpck->prev) {

	 if ( mpck->next ) { 

	    mummy = mpck->next ;
	  
	    mummy0 = mpck ;
	  
	    mpck = mpck->prev ;
	  
	    mpck->next = mummy ;
	  
	    mpck->next->prev = mpck ;
	  
	    	    
            uummy1 = mummy0->ufrag ;           
       
            while ( uummy1 ) {
	  
	       if (uummy1->next) uummy2 = uummy1->next ;
	       else uummy2 = NULL ;
	  
	       free ((void*) uummy1->packet) ;
	       free ((void*) uummy1 ) ;
               
	       uummy1 = uummy2 ;
	    }  
	  
	   free ((void*) mummy0->loctime); 
	   free ((void*) mummy0 ) ;
	 }
	 else {
	    
	    mummy = mpck ;
	    
	    mpck = mpck->prev ;
	    
	    mpck->next = NULL ;
	    
	    uummy1 = mummy->ufrag ;           
       
               while ( uummy1 ) {
	  
	          if (uummy1->next) uummy2 = uummy1->next ;
	          else uummy2 = NULL ;

	          free ((void*) uummy1->packet) ;
		  free ((void*) uummy1 ) ;
		
	          uummy1 = uummy2 ;
	       }
	       free ((void*) mummy->loctime);
	       free ((void*) mummy) ;
	 } 

       }	 
       else if (!mpck->prev) {
	  
	    mummy = mpck ;

	    if ( mpck->next) {
	 
	       mfrag = mpck = mpck->next ;
	       
	       mfrag->prev = NULL ;
	  
	       uummy1 = mummy->ufrag ;           
       
               while ( uummy1 ) {
	  
	          if (uummy1->next) uummy2 = uummy1->next ;
	          else uummy2 = NULL ;

	          free ((void*) uummy1->packet) ;
		  free ((void*) uummy1 ) ;
		
	          uummy1 = uummy2 ;
	       }
	       free ((void*) mummy->loctime);
	       free ((void*) mummy) ;
	    }
	    else {
	       
	         uummy1 = mummy->ufrag ;           
       
                 while ( uummy1 ) {
	  
	               if (uummy1->next) uummy2 = uummy1->next ;
	               else uummy2 = NULL ;

	               free ((void*) uummy1->packet) ;
		       free ((void*) uummy1 ) ;
		
	               uummy1 = uummy2 ;
	         }
	         free ((void*) mummy->loctime);
	         free ((void*) mummy) ;
	         mfrag = mpck = NULL;

	    }
	  
       }
		 
       return (head);
     }
     else { 

timeout_test:	 

	     if ( loctime->tm_sec > mpck->loctime->tm_sec ){ 
                if ( (loctime->tm_sec - mpck->loctime->tm_sec) > 25 ) goto go ;
             }
	     else if ( loctime->tm_sec < mpck->loctime->tm_sec ) {
	      if ((loctime->tm_sec+60-mpck->loctime->tm_sec) > 25 ) goto go ;
	     }

     }
     
     if (mpck) mpck = mpck->next ;
     
   }
   

   return (NULL);   
   
}


      
  

void *thread_read ( void *in)
{

 struct sockaddr_ll scka ;  
   
 struct HEADER *preheader ,*hdummy ;  

 unsigned char head[IPV4HEADER+1] ;
   
 unsigned char *data ;

 unsigned short int version = 0 ;
   
 int msg ,length ,header ,offset ;  

 const int FLUSH = 1 ;  
   
 while (1) {   

    
   if ((msg=recvfrom ( Fd.fd ,head ,1 ,
		           MSG_TRUNC|MSG_PEEK|MSG_DONTWAIT ,NULL,NULL)) < 0){ 
      
      if ( errno==EAGAIN ) { 

	 preheader = ip_assembly (NULL,FLUSH) ;
         
	 goto flush1;
      } 
      
      fprintf ( stdout,"\nERROR :packet reading error (1)\n\n " );
       
      exit(0);
         
   }

   if ( msg >= IPV4HEADER) {  
    
    version = 0 ;
    version = head[0] >> 4 ;
    
    if ( (version != 4) && ( version != 6 ) ) goto clear ; 

    if ( recvfrom ( Fd.fd ,head ,8 ,MSG_PEEK ,NULL ,NULL) < 0 ) { 
      
      fprintf ( stdout,"\nERROR :packet reading error (1)\n\n " );
       
      exit(0);
    }    
      
    if ( version == 4 ) {   

      header = 0 ; 
      header = ( head[0] & 0x0F  ) * 4  ; 
       
      length   = 0 ;
      length   = head[2] ;	      
      length <<= 8 ;	
      length  += head[3] ;
    
      offset   = 0 ; 
      offset   = head[6] & 0x1F ;
      offset <<= 8 ;
      offset  += head[7] ;
       
    }else if ( version == 6 ) {
    
      header = IPV6HEADER ; 
       
      length   = 0 ;
      length   = head[4] ;
      length <<= 8 ;
      length  += head[5] ;
      length  += IPV6HEADER ;
    
      if ( head[6] == IPPROTO_FRAGMENT ) {
	    
       if ( msg >= IPV6HEADER+IPV6FRAG ) {  
	 
        if ( recvfrom (Fd.fd ,head ,
		           IPV6HEADER+IPV6FRAG/2 ,MSG_PEEK ,NULL ,NULL) < 0) { 
      
           fprintf ( stdout,"\nERROR :packet reading error (1)\n\n " );
       
           exit(0);
        }
      
        offset    = head[IPV6HEADER+2] ;
        offset  <<= 8 ;
        offset   += head[IPV6HEADER+3] ;
        offset  >>= 3 ;
       
       }else goto clear ; 
    
      }
      else offset = 0 ;
    }
      
    if ( length != msg ) {
   
       if ( !strncmp ( Fd.interface_name , "eth" ,3 ) ) { 
	  if ( msg == ETHMINLEN ) msg = length ;
       }
       else goto clear ;
    }
      
    if (version == 4 ) {  

       if ( header < IPV4HEADER )              goto clear ;         
       
       if ( !offset && (length < header) )     goto clear ;  
       
       if ( offset && (length < header + 1 ) ) goto clear ;  
    }
    else if ( version == 6 ) {
       
       if ( !offset && (length < IPV6HEADER ) )     goto clear ;
       
       if ( offset && (length < IPV6HEADER + IPV6FRAG + 1 ) ) goto clear ;
    }  
       
       
    if (( data = (char*) malloc (msg+1)) == NULL ) {
              fprintf ( stdout ,"\nERROR :No space for packet data\n\n") ;
              exit (1) ;
    }

    if ( read ( Fd.fd ,data ,msg ) < 0 ){
    
      fprintf (stdout ,"\nERROR :packet reading error (2)\n\n") ;
      exit (0);
    }

    preheader = ip_assembly (data,0) ; 

    free ((void*) data); 
flush1:    
    if ( preheader ) { 

       pthread_mutex_lock (&mutex); 

       while  (READWRITE) pthread_cond_wait (&reading ,&mutex);
       
       if ( Start ) {	       

	       hdummy = Start ;	       
	    
	       while ( hdummy->next )  hdummy = hdummy->next;
	  
               hdummy->next = preheader ;
    
               hdummy->next->next  = NULL ; 
	     
       }
       else { Start = preheader ;Start->next = NULL ; }

       READWRITE = 1 ;
       
       pthread_cond_signal (&writing); 
       
       pthread_mutex_unlock (&mutex);
    }

   }
   else {

clear:      
        if ( read ( Fd.fd , head ,msg ) < 0){
    
           fprintf (stdout ,"\nERROR :packet reading error (2)\n\n") ;
           exit (0);
        }
   }
    
 
 }

}



void alarm4 ( FILE *out2 )
{
  struct OVERWRITE *over ;
  unsigned int length ,count ,fill ,flag ;
  unsigned short int *flags_s_value = Flags_S_value ; 
  char *tmp ;

        fprintf (Out1 ,"\n") ;
   
        if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( Out1,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( Out1 ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
	}
   
        fprintf ( Out1,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]) ;
   
   
        fprintf ( Out1,"\nSource       Address %u.%u.%u.%u\n" ,
	                                              Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] );
      
        fprintf ( Out1,"Destination  Address %u.%u.%u.%u\n\n" ,
	                                              Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
	                                              Start->packet_data[19] );
        fprintf (stdout ,"\n") ;	
   
        if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( stdout,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( stdout ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
	}


		  
        fprintf ( stdout,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]);
   		  
		  
        fprintf ( stdout,"\nSource       Address %u.%u.%u.%u\n\r" ,
	                                              Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] );
      
        fprintf ( stdout,"Destination  Address %u.%u.%u.%u\n\n\r" ,
	                                              Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
	                                              Start->packet_data[19] );
        fprintf (out2 ,"\n") ;
   
        if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( out2,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( out2 ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
	}

	
		  
        fprintf ( out2,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]);
   	
		  
		  
	fprintf ( out2   ,"\nSource       Address %u.%u.%u.%u\n" ,
	                                              Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] );
      
        fprintf ( out2   ,"Destination  Address %u.%u.%u.%u\n\n" ,
	                                              Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
		                                      Start->packet_data[19] );
	

	tmp = Start->packet_data  ; 
        length = Start->packet_length  ; 
      
        while ( length ) {
	
	   for (count = 1 ;count <=4 ;count ++ ) {
	      
	       flag = 1  ;
	      
	       if (! length ) break ;
	      
	       fprintf ( out2 ,"%2X  " ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
	       
	       fprintf ( Out1 ,"%2X  " ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
	   
	       length-- ;flag = 0 ;
	   }

	   if ( flag ) {
	      
	      fill = count-- ;
	      
	      for ( count = fill ; count <= 4 ; count++ ) {
	          
		  fprintf ( out2 ,"    " );
	          fprintf ( Out1 ,"    " );
	      }
	   }
	   else fill = 4 ;
	   
	   fprintf (out2 ,"     ");
	   fprintf (Out1 ,"     ");
	
	   length += fill ;
	
	   for (count = 1;count <=4 ;count++ ) {
	      
	       if (! length) break ;
	      
	       if ( isprint ( *(tmp + Start->packet_length - length)) ) {
		  
		  fprintf ( out2 ,"%c" ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
		  fprintf ( Out1 ,"%c" ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
	       }
	       else { fprintf (out2 ,"*") ;fprintf (Out1 ,"*");}
			     
	       length--;
			     
	   }
	          
	   fprintf (out2 ,"\n");
	   fprintf (Out1 ,"\n");
	      
        }


     if (Start->overwrite) {

fprintf(Out1,"\n-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-\n\n"); 
fprintf(out2,"\n-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-\n\n"); 
		
      over = Start->overwrite ;	
	
      while ( over ) {	

fprintf(Out1  ,"\n---------------------OVERWRITE---------------------\n\n"); 
fprintf(out2  ,"\n---------------------OVERWRITE---------------------\n\n"); 
fprintf(stdout,"\n---------------------OVERWRITE---------------------\n\n"); 
	
	 
	tmp = over->packet_data ;
	length = over->length  ;


        while ( length ) {
	
	   for (count = 1 ;count <=4 ;count ++ ) {
	      
	       flag = 1  ;
	      
	       if (! length ) break ;
	      
	       fprintf ( out2 ,"%2X  " ,
		       (unsigned char)*(tmp + over->length - length) );
	       
	       fprintf ( Out1 ,"%2X  " ,
		       (unsigned char)*(tmp + over->length - length) );
	   
	       length-- ;flag = 0 ;
	   }

	   if ( flag ) {
	      
	      fill = count-- ;
	      
	      for ( count = fill ; count <= 4 ; count++ ) {
	          
		  fprintf ( out2 ,"    " );
	          fprintf ( Out1 ,"    " );
	      }
	   }
	   else fill = 4 ;
	   
	   fprintf (out2 ,"     ");
	   fprintf (Out1 ,"     ");
	
	   length += fill ;
	
	   for (count = 1;count <=4 ;count++ ) {
	      
	       if (! length) break ;
	      
	       if ( isprint ( *(tmp + over->length - length)) ) {
		  
		  fprintf ( out2 ,"%c" ,
		       (unsigned char)*(tmp + over->length - length) );
		  fprintf ( Out1 ,"%c" ,
		       (unsigned char)*(tmp + over->length - length) );
	       }
	       else { fprintf (out2 ,"*") ;fprintf (Out1 ,"*");}
			     
	       length--;
			     
	   }
	          
	   fprintf (out2 ,"\n");
	   fprintf (Out1 ,"\n");
	      
        }

	fprintf ( Out1 ,"\n\nOffset : %d\n\n"   ,over->offset );
	fprintf ( out2 ,"\n\nOffset : %d\n\n"   ,over->offset );   
	fprintf ( stdout ,"\n\nOffset : %d\n\n" ,over->offset );

        over = over->next ;
      }
	
	
     }
   
     fflush (out2) ;
     fclose (out2) ;
        
     fflush (Out1);
	
     fprintf (stdout,"\a\a");
     fflush (stdout);

	
}

void alarm6 ( FILE *out2 )
{
  struct OVERWRITE *over ;
  unsigned int length ,count ,fill ,flag ;
  unsigned short int  *flags_s_value = Flags_S_value ;
  char *tmp ; 

        fprintf (Out1 ,"\n") ;
   
        if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( Out1,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( Out1 ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
	}


   
      fprintf ( Out1,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]);
   
   
      fprintf ( Out1,"\nSource      Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n",
	                                              Start->packet_data[8]  ,
	                                              Start->packet_data[9]  ,
	                                              Start->packet_data[10] ,
	                                              Start->packet_data[11] ,
                                                      Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] ,
                                                      Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
	                                              Start->packet_data[19] ,
                                                      Start->packet_data[20] ,
	                                              Start->packet_data[21] ,
	                                              Start->packet_data[22] ,
	                                              Start->packet_data[23] );

      fprintf ( Out1,"\nDestination Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n\n",
	                                              Start->packet_data[24]  ,
	                                              Start->packet_data[25]  ,
	                                              Start->packet_data[26] ,
	                                              Start->packet_data[27] ,
                                                      Start->packet_data[28] ,
	                                              Start->packet_data[29] ,
	                                              Start->packet_data[30] ,
	                                              Start->packet_data[31] ,
                                                      Start->packet_data[32] ,
	                                              Start->packet_data[33] ,
	                                              Start->packet_data[34] ,
	                                              Start->packet_data[35] ,
                                                      Start->packet_data[36] ,
	                                              Start->packet_data[37] ,
	                                              Start->packet_data[38] ,
	                                              Start->packet_data[39] );
        fprintf (stdout ,"\n") ;
     
        if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( stdout,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( stdout ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
	}

		
		
		
      fprintf ( stdout,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]);
   		
		
      fprintf ( stdout,"\nSource      Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n\r",
	                                              Start->packet_data[8]  ,
	                                              Start->packet_data[9]  ,
	                                              Start->packet_data[10] ,
	                                              Start->packet_data[11] ,
                                                      Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] ,
                                                      Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
	                                              Start->packet_data[19] ,
                                                      Start->packet_data[20] ,
	                                              Start->packet_data[21] ,
	                                              Start->packet_data[22] ,
	                                              Start->packet_data[23] );

      fprintf ( stdout,"\nDestination Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n\n\r",
	                                              Start->packet_data[24]  ,
	                                              Start->packet_data[25]  ,
	                                              Start->packet_data[26] ,
	                                              Start->packet_data[27] ,
                                                      Start->packet_data[28] ,
	                                              Start->packet_data[29] ,
	                                              Start->packet_data[30] ,
	                                              Start->packet_data[31] ,
                                                      Start->packet_data[32] ,
	                                              Start->packet_data[33] ,
	                                              Start->packet_data[34] ,
	                                              Start->packet_data[35] ,
                                                      Start->packet_data[36] ,
	                                              Start->packet_data[37] ,
	                                              Start->packet_data[38] ,
	                                              Start->packet_data[39] );
        fprintf ( out2,"\n") ;
   
        if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( out2,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( out2 ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
	}


		
	 fprintf ( out2,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]);
   
		
         fprintf ( out2,"\nSource      Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n",
	                                              Start->packet_data[8]  ,
	                                              Start->packet_data[9]  ,
	                                              Start->packet_data[10] ,
	                                              Start->packet_data[11] ,
                                                      Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] ,
                                                      Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
	                                              Start->packet_data[19] ,
                                                      Start->packet_data[20] ,
	                                              Start->packet_data[21] ,
	                                              Start->packet_data[22] ,
	                                              Start->packet_data[23] );

      fprintf ( out2,"\nDestination Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n\n",
	                                              Start->packet_data[24]  ,
	                                              Start->packet_data[25]  ,
	                                              Start->packet_data[26] ,
	                                              Start->packet_data[27] ,
                                                      Start->packet_data[28] ,
	                                              Start->packet_data[29] ,
	                                              Start->packet_data[30] ,
	                                              Start->packet_data[31] ,
                                                      Start->packet_data[32] ,
	                                              Start->packet_data[33] ,
	                                              Start->packet_data[34] ,
	                                              Start->packet_data[35] ,
                                                      Start->packet_data[36] ,
	                                              Start->packet_data[37] ,
	                                              Start->packet_data[38] ,
	                                              Start->packet_data[39] );

   
	

	tmp = Start->packet_data  ; 
        length = Start->packet_length  ; 
      
        while ( length ) {
	
	   for (count = 1 ;count <=4 ;count ++ ) {
	      
	       flag = 1 ;
	      
	       if (! length ) break ;
	      
	       fprintf ( out2 ,"%2X  " ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
	       
	       fprintf ( Out1 ,"%2X  " ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
	   
	       length-- ;flag = 0;
	   }

	   if ( flag ) {
	      
	      fill = count-- ;
	      
	      for ( count = fill ; count <= 4 ; count++ ) {
	          
		  fprintf ( out2 ,"    " );
	          fprintf ( Out1 ,"    " );
	      }
	   }
	   else fill = 4 ;
	   
	   fprintf (out2 ,"     ");
	   fprintf (Out1 ,"     ");
	
	   length += fill  ;
	
	   for (count = 1;count <=4 ;count++ ) {
	      
	       if (! length) break ;
	      
	       if ( isprint ( *(tmp + Start->packet_length - length)) ) {
		  
		  fprintf ( out2 ,"%c" ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
		  fprintf ( Out1 ,"%c" ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
	       }
	       else { fprintf (out2 ,"*") ;fprintf (Out1 ,"*");}
			     
	       length--;
			     
	   }
	          
	   fprintf (out2 ,"\n");
	   fprintf (Out1 ,"\n");
	      
        }


     if (Start->overwrite) {

fprintf(Out1,"\n-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-\n\n"); 
fprintf(out2,"\n-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-OVERWRITE-\n\n"); 
		
      over = Start->overwrite ;	
	
      while ( over ) {	

fprintf(Out1  ,"\n---------------------OVERWRITE---------------------\n\n"); 
fprintf(out2  ,"\n---------------------OVERWRITE---------------------\n\n"); 
fprintf(stdout,"\n---------------------OVERWRITE---------------------\n\n"); 
	
	 
	tmp = over->packet_data ;
	length = over->length  ;


        while ( length ) {
	
	   for (count = 1 ;count <=4 ;count ++ ) {
	      
	       flag = 1  ;
	      
	       if (! length ) break ;
	      
	       fprintf ( out2 ,"%2X  " ,
		       (unsigned char)*(tmp + over->length - length) );
	       
	       fprintf ( Out1 ,"%2X  " ,
		       (unsigned char)*(tmp + over->length - length) );
	   
	       length-- ;flag = 0 ;
	   }

	   if ( flag ) {
	      
	      fill = count-- ;
	      
	      for ( count = fill ; count <= 4 ; count++ ) {
	          
		  fprintf ( out2 ,"    " );
	          fprintf ( Out1 ,"    " );
	      }
	   }
	   else fill = 4 ;
	   
	   fprintf (out2 ,"     ");
	   fprintf (Out1 ,"     ");
	
	   length += fill ;
	
	   for (count = 1;count <=4 ;count++ ) {
	      
	       if (! length) break ;
	      
	       if ( isprint ( *(tmp + over->length - length)) ) {
		  
		  fprintf ( out2 ,"%c" ,
		       (unsigned char)*(tmp + over->length - length) );
		  fprintf ( Out1 ,"%c" ,
		       (unsigned char)*(tmp + over->length - length) );
	       }
	       else { fprintf (out2 ,"*") ;fprintf (Out1 ,"*");}
			     
	       length--;
			     
	   }
	          
	   fprintf (out2 ,"\n");
	   fprintf (Out1 ,"\n");
	      
        }

	fprintf ( Out1 ,"\n\nOffset : %d\n\n"   ,over->offset );
	fprintf ( out2 ,"\n\nOffset : %d\n\n"   ,over->offset );   
	fprintf ( stdout ,"\n\nOffset : %d\n\n" ,over->offset );

        over = over->next ;
      }
	
	
     }   
   
     fflush (out2) ;
     fclose (out2) ;
        
     fflush (Out1);
	
     fprintf (stdout,"\a\a");
     fflush (stdout);

	
}

void *shift6 ( unsigned int  part1 ,unsigned int part2 ,
	       unsigned int  part3 ,unsigned int part4 ,
	       unsigned short int slip )
{
   
   static unsigned int result[4] ;
   unsigned int test = 0x00000001 ;
   unsigned int on   = 0x80000000 ;
   unsigned int count ,rt1 ,rt2 ,rt3 ;
   
   if ( slip ) {
   
    for ( count = 1 ; count <= slip ; count++ ) {
  
       rt1 = part1 & test ;
       rt2 = part2 & test ;
       rt3 = part3 & test ;
            
       part1 >>= 1 ; part2 >>= 1 ;part3 >>= 1 ;part4 >>= 1 ;
      
       if ( rt1 ) part2 |= on ;
       if ( rt2 ) part3 |= on ;
       if ( rt3 ) part4 |= on ;
       
    }
   
    result[0] = part1 ;
    result[1] = part2 ;
    result[2] = part3 ;
    result[3] = part4 ;
   
    return ( (void*) result );
   }
   return (NULL) ;
}

int test6 ( void *first ,void* last ,unsigned short int slip ) {
   
                                  
    unsigned int head[4] ;
    unsigned int tail[4] ;
    void *result1 ,*result2 ;
   
   
    memcpy ( (void*)&head[0] ,(void*)&first ,16 );
    memcpy ( (void*)&tail[0] ,(void*)&last  ,16 );

   
    result1 = shift6 ( head[0] ,head[1] ,head[2] ,head[3] ,128 - slip ) ;
    result2 = shift6 ( tail[0] ,tail[1] ,tail[3] ,tail[3] ,128 - slip ) ;
   
    if ( result1 && result2 ) {
   
       memcpy ( (void*)&head[0] ,result1  ,16 );
       memcpy ( (void*)&tail[0] ,result2  ,16 );
   
      if ( ((head[0] & tail[0]) == tail[0] ) &&
           ((head[1] & tail[1]) == tail[1] ) && 
	   ((head[2] & tail[2]) == tail[2] ) &&
	   ((head[3] & tail[3]) == tail[3] )
	   
	   )   return (1) ;
      
     return (0);
    }
    return (1);
}

void *thread_write ( void *in ) 
{  
 
 struct OVERWRITE *over ,*overtmp ;  
 struct HEADER *dummy ; 
 struct FLAGS  *stflags = Flags ;
 struct PORT   *wport ;
 unsigned short int Sport ,Dport ,count ,fill ,b_flag ,
                    Hlen ,flags ,Nulludp ,Proto ,version ;
 unsigned short int *flags_s_value = Flags_S_value ;
 unsigned int  length ,netaddr ,localaddr;
       
 unsigned int netaddress6[4] ; 
   
 char *tmp ; 

 int fabio=0 ;  
   
 time_t curtime ;
 struct tm *loctime ;

 FILE *out2  ;  
 static unsigned int filenmbr = 1 ;  
 char nmbr[100] ;  

 struct OPTIONS *options = NULL ;    

 if ( mkdir ( Directory , 0666 & ~umask(0) ) ) {

    	fprintf (stdout ,"\n\nERROR : Can't make directory \n\n");
	exit (1);
 }
   
 if ( (Out1 = fopen ( stcat ( Directory ,"PKDATA" ,1 ) ,"w" )) == NULL ) {
        fprintf (stdout ,"\n\nERROR : Can't open PKDATA file \n\n");
	exit (1);
 }
   
 while ( 1 ) {  
    
  OUTGOING_FLAG = L_FLAGS = 0 ;  
    
  tzset ();  
    
  curtime = time (NULL);  
    
  loctime = localtime (&curtime);    

  pthread_mutex_lock (&mutex) ;
    
  while (!READWRITE) pthread_cond_wait (&writing ,&mutex);  
    
  if (Start) {   
    
    version = Start->packet_data[0] >> 4  ;      
      
    if ( version == 4 ) { 
      
     Hlen = (Start->packet_data[0] & 0x0F  ) * 4  ; 
    
     Proto = Start->packet_data[9] ;
       
     fabio = 1 ;  
    
    }    
    else {

          Hlen = IPV6HEADER ;
       
          if ( Start->packet_data[6] != IPPROTO_TCP && 
	       Start->packet_data[6] != IPPROTO_UDP ) {

           while (1) {

	      if ( Start->packet_data[Hlen] == IPPROTO_NONE ) break ;
	      
	      if ( Start->packet_data[Hlen] != IPPROTO_TCP &&
		   Start->packet_data[Hlen] != IPPROTO_UDP )  
		   
		   Hlen += Start->packet_data[Hlen+1] ;
	     
	      else { 
		 
		 Proto = Start->packet_data[Hlen] ;
		 Hlen += Start->packet_data[Hlen+1] ;
		 fabio = 1;
		 break ;
	      }
	   }
	  }
          else {Proto = Start->packet_data[6] ;fabio = 1 ;}
    }
    
   if (fabio) { 
	   
      
    switch ( Proto ) {

       
    case IPPROTO_TCP :fprintf (Out1,"\nProtocol TCP\n\n");
	             
	             Sport = 0 ;      
                     Sport = Start->packet_data[Hlen+0] ;
                     Sport <<= 8 ;
                     Sport += Start->packet_data[Hlen+1] ;
	             Dport = 0 ; 
                     Dport = Start->packet_data[Hlen+2] ;
                     Dport <<= 8 ;
                     Dport += Start->packet_data[Hlen+3] ;     
                     
		     L_FLAGS = flags = Start->packet_data[Hlen+13] & 0x3F ;

		     if ( version == 4 ) { 
			    
			      netaddr =   0 ;
			      netaddr =   Start->packet_data[12];
			      netaddr <<= 8 ;
	                      netaddr +=  Start->packet_data[13];
			      netaddr <<= 8 ;
	                      netaddr +=  Start->packet_data[14];
			      netaddr <<= 8 ;
	                      netaddr +=  Start->packet_data[15];
			     
			      localaddr = htonl ( iddr->s_addr ) ;
			
			      if ( localaddr == netaddr ) OUTGOING_FLAG = 1;
			
			      if ( (localaddr == netaddr) && 
				   IGNORE_FLAG ) goto stop ;
		     }
       		     else {

			    netaddress6[0] = 
			    netaddress6[1] = 
			    netaddress6[2] =
			    netaddress6[3] = 0 ; 
			     
		            memcpy ( (void*)&netaddress6[0] ,
				     (void*)((Start->packet_data)+8), 16);
			   
			    memcpy ( (void*)&(iddr6->s6_addr[0]) ,
				     (void*)(Start->packet_data+24) ,16 );
			
			    for ( count = 0 ;count <=4 ;) { 
			     
			        if ( count == 4 ) OUTGOING_FLAG = 1 ;
			       
			        if ((count == 4) && IGNORE_FLAG ) goto stop ;
			       
			        if ( iddr6->s6_addr32[count] ==
				     netaddress6[count] ) count++ ; 
                                else break ;
			    }
			
                     }
                           
	             options = Options ;
	
	             if ( options ) {
	     
	                while ( options ) {
			   
			 if ( options->opt == INTCP ) { 
			      
			  if ( version == 4 ) { 
			    
			   if ( !options->rport) {  
			    
			    if ((((Dport >= options->sport)  &&
				(Dport <= options->eport))  ||   
				((Dport == options->sport)  &&
				 (options->eport == 0 )))   &&    
                             ((((options->myaddress >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))
			       ) goto stop ;
			   }
			   else {
			    
			        wport = options->rport ;
			      
			        while ( wport ) {
				   
				   if ( Dport == wport->port ) break;
				
				   wport = wport->next ;
				   
				}
			        if ( wport &&
				((((options->myaddress >> (32-options->mask)) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))
			        ) goto stop ;
			   }
			      	   
				  
			  }/* if version == 4 */ 
			  else {

			    if ( test6 ( (void*) &options->myaddress6[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask ) 
				       )
			    {
			
			     if ( !options->rport ) {   
				    
			     
			      if ((((Dport >= options->sport)  &&
				    (Dport <= options->eport))  ||   
				   ((Dport == options->sport)  &&
				   (options->eport == 0 ))) ) goto stop ;
			     }
			     else {
				  wport = options->rport ;
			      
			          while ( wport ) {
				   
				        if ( Dport == wport->port ) break;
				
				        wport = wport->next ;
				   
				  }
				  if (wport) goto stop ;
			     }
			       
			    }
			     
			  }
			    
			 }/*if INTCP */ 			   
			 else if ( options->opt == OUTTCP ) { 
			      
			  if ( version == 4 ) { 
			    
			   if ( !options->rport) {  
			    
			    if ((((Dport >= options->sport)  &&
				(Dport <= options->eport))  ||   
				((Dport == options->sport)  &&
				 (options->eport == 0 )))   &&    
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))
			       ) goto stop ;
			   }
			   else {
			    
			        wport = options->rport ;
			      
			        while ( wport ) {
				   
				   if ( Dport == wport->port ) break;
				
				   wport = wport->next ;
				   
				}
			        if ( wport &&
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))     
			        ) goto stop ;
			   }
			      	   
				  
			  }/* if version == 4 */ 
			  else {

			    if ( test6 ( (void*) iddr6->s6_addr32[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask )
				 )
			    {
			
			     if ( !options->rport ) {   
				    
			     
			      if ((((Dport >= options->sport)  &&
				    (Dport <= options->eport))  ||   
				   ((Dport == options->sport)  &&
				   (options->eport == 0 ))) ) goto stop ;
			     }
			     else {
				  wport = options->rport ;
			      
			          while ( wport ) {
				   
				        if ( Dport == wport->port ) break;
				
				        wport = wport->next ;
				   
				  }
				  if (wport) goto stop ;
			     }
			       
			    }
			     
			  }
			    
			 }/*if OUTTCP */
			 else if ( options->opt == ALLTCP ) { 
			      
			  if ( version == 4 ) { 
			    
			   if ( !options->rport) {  
			    
			    if ((((Dport >= options->sport)  &&
				(Dport <= options->eport))  ||   
				((Dport == options->sport)  &&
				 (options->eport == 0 )))   &&    
                             (((((options->myaddress >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))) ||
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))))
			       ) goto stop ;
			   }
			   else {
			    
			        wport = options->rport ;
			      
			        while ( wport ) {
				   
				   if ( Dport == wport->port ) break;
				
				   wport = wport->next ;
				   
				}
			        if ( wport &&
				(((((options->myaddress >> (32-options->mask)) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))) ||
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))))     
			        ) goto stop ;
			   }
			      	   
				  
			  }/* if version == 4 */ 
			  else {

			    if ( test6 ( (void*) &options->myaddress6[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask ) ||
				 test6 ( (void*) iddr6->s6_addr32[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask )
				 )
			    {
			
			     if ( !options->rport ) {   
				    
			     
			      if ((((Dport >= options->sport)  &&
				    (Dport <= options->eport))  ||   
				   ((Dport == options->sport)  &&
				   (options->eport == 0 ))) ) goto stop ;
			     }
			     else {
				  wport = options->rport ;
			      
			          while ( wport ) {
				   
				        if ( Dport == wport->port ) break;
				
				        wport = wport->next ;
				   
				  }
				  if (wport) goto stop ;
			     }
			       
				    
			       
			    }
			     
			     
			     
			     
			  }
			   
			    
			 }/* ALLTCP */  
			
			 options = options->next ;
			
			}
			
		     }
	
	          do {

		     if ( (stflags->flags & Flags_value[1]) && (!flags) ) 
	                                               flags |= Flags_value[1];
		     
	             if ( (flags == stflags->flags) ) {
			
			if ( (out2 = fopen ( stcat (Directory ,
			      iitoa ( filenmbr ,nmbr) ,1),"w") ) == NULL ) {
                   fprintf (stdout ,"\n\nERROR : Can't open scan file \n\n");
	                      goto stop ;
                        }
			   
			   
			     
fprintf(Out1,"\n######################################################\n\n");
fprintf(stdout,"\n######################################################\n\r");
fprintf(out2  ,"\n######################################################\n\n");

                        fprintf ( stdout,"\nProtocol TCP\n\n\r");
	                fprintf ( Out1   ,"\nProtocol TCP\n\n"); 
			fprintf ( out2   ,"\nProtocol TCP\n\n");
	             
			fprintf ( stdout,"\nInterface  : %s\n\n\r",IFACE);
	                fprintf ( Out1   ,"\nInteface  : %s\n\n",IFACE);
			fprintf ( out2   ,"\nInteface  : %s\n\n",IFACE);
			
			
		        fputs ( asctime(loctime) ,stdout );
	                fputs ( asctime(loctime) ,Out1 );
			fputs ( asctime(loctime) ,out2 );
			
			fprintf (stdout,"\r\n");
			fprintf (Out1,"\n\n");
                        fprintf (out2,"\n\n");
	
	
	                fprintf (stdout ,"SCAN FLAGS = " );
	                fprintf (Out1   ,"SCAN FLAGS = " );
	                fprintf (out2   ,"SCAN FLAGS = " );
	
			count = 1 ;
	
	                while ( Flags_value[count] ){
			   
			  if ( (stflags->flags & Flags_value[count]) ==
			       Flags_value[count] ){
				
	                     fprintf (stdout ,"%s :" ,Flags_name[count]);
	                     fprintf (out2   ,"%s :" ,Flags_name[count]);
			     fprintf (Out1   ,"%s :" ,Flags_name[count]);    
			  }
			  count++;
			}
	
			fprintf (stdout,"\r\n");
			fprintf (Out1,"\r\n");
                        fprintf (out2,"\r\n");
				      

                        fprintf ( stdout,"\nSource      Port %u\r"   ,Sport);
      	                fprintf ( stdout,"\nDestination Port %u\n\r" ,Dport);

			fprintf ( out2   ,"\nSource      Port %u\r"  ,Sport );
      	                fprintf ( out2   ,"\nDestination Port %u\n"  ,Dport );

		        fprintf ( Out1 ,"\nSource      Port %u\r"    ,Sport );
                        fprintf ( Out1 ,"\nDestination Port %u\n\n"  ,Dport );
			
			fprintf (stdout,"\r\n");
			fprintf (Out1,"\r\n");
                        fprintf (out2,"\r\n");
				      

                        fprintf ( stdout,"\nNumber of Fragments %u\n\r" ,
				                          Start->frag_number );
      	                
			fprintf ( out2 ,"\nNumber of Fragments %u\n"  ,
				                          Start->frag_number );

		        fprintf ( Out1 ,"\nNumber of Fragments %u\n\n"    ,
				                          Start->frag_number );
			
			
			if (version == 4) alarm4 (out2);
			else alarm6 (out2);
			
			filenmbr++;
		     }
			
		  stflags = stflags->next ;
		     
		     
		  }while ( stflags );
		
                     stflags = Flags ;            
	
	             break ;
            
	
   case IPPROTO_UDP :fprintf (Out1,"\nProtocol UDP\n\n");
	             Sport = 0 ;      
                     Sport = Start->packet_data[Hlen+0] ;
                     Sport <<= 8 ;
                     Sport += Start->packet_data[Hlen+1] ;
	             Dport = 0 ; 
                     Dport = Start->packet_data[Hlen+2] ;
                     Dport <<= 8 ;
                     Dport += Start->packet_data[Hlen+3] ;     
                     
	             Nulludp   = Start->packet_data[Hlen+4] ;
	             Nulludp <<= 8 ;
	             Nulludp  += Start->packet_data[Hlen+5] ; 

                     L_FLAGS = 0 ;

		     if ( version == 4 ) { 
			    
			      netaddr =   0 ;
			      netaddr =   Start->packet_data[12];
			      netaddr <<= 8 ;
	                      netaddr +=  Start->packet_data[13];
			      netaddr <<= 8 ;
	                      netaddr +=  Start->packet_data[14];
			      netaddr <<= 8 ;
	                      netaddr +=  Start->packet_data[15];
			     
			      localaddr = htonl ( iddr->s_addr ) ;
			
			      if ( localaddr == netaddr ) OUTGOING_FLAG = 1;
			
			      if ( (localaddr == netaddr) && 
				   IGNORE_FLAG ) goto stop ;
		     }
       		     else {

			    netaddress6[0] = 
			    netaddress6[1] = 
			    netaddress6[2] =
			    netaddress6[3] = 0 ; 
			     
		            memcpy ( (void*)&netaddress6[0] ,
				     (void*)((Start->packet_data)+8), 16);
			   
			    memcpy ( (void*)&(iddr6->s6_addr[0]) ,
				     (void*)(Start->packet_data+24) ,16 );
			
			    for ( count = 0 ;count <=4 ;) { 
			     
			        if ( count == 4 ) OUTGOING_FLAG = 1 ;
			       
			        if ((count == 4) && IGNORE_FLAG ) goto stop ;
			       
			        if ( iddr6->s6_addr32[count] ==
				     netaddress6[count] ) count++ ; 
                                else break ;
			    }
			
                     }
                           
	             options = Options ;
	
	             if ( options ) {
	     
	                while ( options ) {
			   
			 if ( options->opt == INUDP ) { 
			      
			  if ( version == 4 ) { 
			    
			   if ( !options->rport) {  
			    
			    if ((((Dport >= options->sport)  &&
				(Dport <= options->eport))  ||   
				((Dport == options->sport)  &&
				 (options->eport == 0 )))   &&    
                             ((((options->myaddress >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))
			       ) goto stop ;
			   }
			   else {
			    
			        wport = options->rport ;
			      
			        while ( wport ) {
				   
				   if ( Dport == wport->port ) break;
				
				   wport = wport->next ;
				   
				}
			        if ( wport &&
				((((options->myaddress >> (32-options->mask)) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))
			        ) goto stop ;
			   }
			      	   
				  
			  }/* if version == 4 */ 
			  else {

			    if ( test6 ( (void*) &options->myaddress6[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask ) 
				       )
			    {
			
			     if ( !options->rport ) {   
				    
			     
			      if ((((Dport >= options->sport)  &&
				    (Dport <= options->eport))  ||   
				   ((Dport == options->sport)  &&
				   (options->eport == 0 ))) ) goto stop ;
			     }
			     else {
				  wport = options->rport ;
			      
			          while ( wport ) {
				   
				        if ( Dport == wport->port ) break;
				
				        wport = wport->next ;
				   
				  }
				  if (wport) goto stop ;
			     }
			       
			    }
			     
			  }
			    
			 }/*if INUDP */ 			   
			 else if ( options->opt == OUTUDP ) { 
			      
			  if ( version == 4 ) { 
			    
			   if ( !options->rport) {  
			    
			    if ((((Dport >= options->sport)  &&
				(Dport <= options->eport))  ||   
				((Dport == options->sport)  &&
				 (options->eport == 0 )))   &&    
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))
			       ) goto stop ;
			   }
			   else {
			    
			        wport = options->rport ;
			      
			        while ( wport ) {
				   
				   if ( Dport == wport->port ) break;
				
				   wport = wport->next ;
				   
				}
			        if ( wport &&
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask)))     
			        ) goto stop ;
			   }
			      	   
				  
			  }/* if version == 4 */ 
			  else {

			    if ( test6 ( (void*) iddr6->s6_addr32[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask )
				 )
			    {
			
			     if ( !options->rport ) {   
				    
			     
			      if ((((Dport >= options->sport)  &&
				    (Dport <= options->eport))  ||   
				   ((Dport == options->sport)  &&
				   (options->eport == 0 ))) ) goto stop ;
			     }
			     else {
				  wport = options->rport ;
			      
			          while ( wport ) {
				   
				        if ( Dport == wport->port ) break;
				
				        wport = wport->next ;
				   
				  }
				  if (wport) goto stop ;
			     }
			       
			    }
			     
			  }
			    
			 }/*if OUTUDP */
			 else if ( options->opt == ALLUDP ) { 
			      
			  if ( version == 4 ) { 
			    
			   if ( !options->rport) {  
			    
			    if ((((Dport >= options->sport)  &&
				(Dport <= options->eport))  ||   
				((Dport == options->sport)  &&
				 (options->eport == 0 )))   &&    
                             (((((options->myaddress >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))) ||
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))))
			       ) goto stop ;
			   }
			   else {
			    
			        wport = options->rport ;
			      
			        while ( wport ) {
				   
				   if ( Dport == wport->port ) break;
				
				   wport = wport->next ;
				   
				}
			        if ( wport &&
				(((((options->myaddress >> (32-options->mask)) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))) ||
				((((localaddr >> (32-options->mask) ) &
				(netaddr            >> (32-options->mask)))) ==
			        (netaddr >> (32-options->mask))))     
			        ) goto stop ;
			   }
			      	   
				  
			  }/* if version == 4 */ 
			  else {

			    if ( test6 ( (void*) &options->myaddress6[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask ) ||
				 test6 ( (void*) iddr6->s6_addr32[0] ,
					 (void*) &netaddress6[0] ,
					 options->mask )
				 )
			    {
			
			     if ( !options->rport ) {   
				    
			     
			      if ((((Dport >= options->sport)  &&
				    (Dport <= options->eport))  ||   
				   ((Dport == options->sport)  &&
				   (options->eport == 0 ))) ) goto stop ;
			     }
			     else {
				  wport = options->rport ;
			      
			          while ( wport ) {
				   
				        if ( Dport == wport->port ) break;
				
				        wport = wport->next ;
				   
				  }
				  if (wport) goto stop ;
			     }
			       
				    
			       
			    }
			     
			     
			     
			     
			  }
			   
			    
			 }/* ALLUDP */  
			
			 options = options->next ;
			
			}
			
		     }
       
	          do {
		     
		     if ( (stflags->flags & Flags_value[2]) &&
			  ( Nulludp == 0x08 ) ) flags = Flags_value[2];
		     
		     else if ( stflags->flags & Flags_value[3] ) 
		                                       flags = Flags_value[3];
		     else flags = 0 ;
		     
	             if ((flags==Flags_value[2])||(flags==Flags_value[3])){ 

			if ( (out2 = fopen ( stcat (Directory ,
			      iitoa ( filenmbr ,nmbr) ,1),"w") ) == NULL ) {
                   fprintf (stdout ,"\n\nERROR : Can't open scan file \n\n");
	                      goto stop ;
                        }
			
fprintf(Out1,"\n######################################################\n\n");
fprintf(stdout,"\n######################################################\n\r");
fprintf(out2  ,"\n######################################################\n\r");

                        fprintf ( stdout,"\nProtocol UDP\n\n\r");
	                fprintf ( Out1   ,"\nProtocol UDP\n\n");
			fprintf ( out2   ,"\nProtocol UDP\n\n");
	             
			
			fprintf ( stdout,"\nInterface  : %s\n\n\r",IFACE);
	                fprintf ( Out1   ,"\nInteface  : %s\n\n",IFACE);
			fprintf ( out2   ,"\nInteface  : %s\n\n",IFACE);
			
			
		        fputs ( asctime(loctime) ,stdout );
	                fputs ( asctime(loctime) ,Out1 );
			fputs ( asctime(loctime) ,out2 );
			
		        fprintf (stdout,"\r\n");
			fprintf (Out1  ,"\n\n");
			fprintf (out2  ,"\n\n");

			
			
	                fprintf (stdout ,"REASON = " );
	                fprintf (Out1   ,"REASON = " );
	                fprintf (out2   ,"REASON = " );
	
			count = 1 ;
			
	                while ( Flags_value[count] ) {
			   
			  if ( flags == Flags_value[count] ){
				
	                     fprintf (stdout ,"%s :" ,Flags_name[count]);
	                     fprintf (out2   ,"%s :" ,Flags_name[count]);
			     fprintf (Out1   ,"%s :" ,Flags_name[count]);    
			  }
			  count++;
			}
	
			fprintf (stdout,"\r\n");
			fprintf (Out1,"\n\n");
                        fprintf (out2,"\n\n");
			
			
			
                        fprintf ( stdout,"\nSource      Port %u\r"   ,Sport );
      	                fprintf ( stdout,"\nDestination Port %u\n\r" ,Dport );
			
			fprintf ( out2  ,"\nSource      Port %u\r" ,Sport);
      	                fprintf ( out2  ,"\nDestination Port %u\n" ,Dport );
		
			fprintf ( Out1 ,"\nSource      Port %u\r"   ,Sport );
                        fprintf ( Out1 ,"\nDestination Port %u\n\n" ,Dport ) ;
			
			fprintf (stdout,"\r\n");
			fprintf (Out1,"\r\n");
                        fprintf (out2,"\r\n");
				      

                        fprintf ( stdout,"\nNumber of Fragments %u\n\r" ,
				                          Start->frag_number );
      	                
			fprintf ( out2 ,"\nNumber of Fragments %u\n"  ,
				                          Start->frag_number );

		        fprintf ( Out1 ,"\nNumber of Fragments %u\n"    ,
				                          Start->frag_number );
			
		        if (version == 4 ) alarm4(out2);
			else alarm6(out2);
			
			filenmbr++;
		     }
		     
	             
		   stflags = stflags->next ;   
		      
		   }while ( stflags );
		    
	             stflags = Flags ;
	
	             break ;

     } 
   }   
      
stop:      
   
     fabio = 0 ; 

     fprintf ( Out1 ,"\nInteface........... %s\r\n",IFACE); 
     fprintf ( Out1 ,"\nSource         Port %u\r"   ,Sport );
     fprintf ( Out1 ,"\nDestination    Port %u\r\n" ,Dport ) ;
     fprintf ( Out1 ,"\nNumber of Fragments %u\r\n" ,Start->frag_number ) ;

      
     if (version == 4 ) { 	

      fprintf ( Out1 ,"\n");	
	
      if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( Out1,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( Out1 ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
      }
	
      fprintf ( Out1,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]);
	
	
      fprintf ( Out1,"\nSource       Address %u.%u.%u.%u\n" ,
	                                              Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] );
      
      fprintf ( Out1,"Destination  Address %u.%u.%u.%u\n\n" ,
    	                                              Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
                                                      Start->packet_data[19] );
     }
     else { 
      
      fprintf ( Out1 ,"\n");	
	
      if ( L_FLAGS ) {
	 
	   flags_s_value = Flags_S_value ;count = 0 ;
	   
	   fprintf ( Out1,"Flag          Status ");

	   while ( *flags_s_value ) {
	    
	         if (  L_FLAGS & *flags_s_value )
					                                 
	            fprintf  ( Out1 ,"%s" ,Flags_S_name[count]);
	      
	         flags_s_value++;count++;
	   }
	   
      }


      
      fprintf ( Out1,"\nPacket     direction :%s\n",
		                               Direction_name[OUTGOING_FLAG]);
      	
	
      fprintf ( Out1,"\nSource      Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n",
	                                              Start->packet_data[8]  ,
	                                              Start->packet_data[9]  ,
	                                              Start->packet_data[10] ,
	                                              Start->packet_data[11] ,
                                                      Start->packet_data[12] ,
	                                              Start->packet_data[13] ,
	                                              Start->packet_data[14] ,
	                                              Start->packet_data[15] ,
                                                      Start->packet_data[16] ,
	                                              Start->packet_data[17] ,
	                                              Start->packet_data[18] ,
	                                              Start->packet_data[19] ,
                                                      Start->packet_data[20] ,
	                                              Start->packet_data[21] ,
	                                              Start->packet_data[22] ,
	                                              Start->packet_data[23] );

      fprintf ( Out1,"\nDestination Address \
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\
%02X%02X:%02X%02X:%02X%02X:%02X%02X:\n\n",
	                                              Start->packet_data[24]  ,
	                                              Start->packet_data[25]  ,
	                                              Start->packet_data[26] ,
	                                              Start->packet_data[27] ,
                                                      Start->packet_data[28] ,
	                                              Start->packet_data[29] ,
	                                              Start->packet_data[30] ,
	                                              Start->packet_data[31] ,
                                                      Start->packet_data[32] ,
	                                              Start->packet_data[33] ,
	                                              Start->packet_data[34] ,
	                                              Start->packet_data[35] ,
                                                      Start->packet_data[36] ,
	                                              Start->packet_data[37] ,
	                                              Start->packet_data[38] ,
	                                              Start->packet_data[39] );

	
	
     }
      
     tmp = Start->packet_data  ; 
     length = Start->packet_length  ; 
      
     while ( length ) {
	
	   for (count = 1 ;count <=4 ;count ++ ) {
	      
	       b_flag = 1 ;
	      
	       if (! length ) break ;
	      
	       fprintf ( Out1 ,"%2X  " ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
	   
	       length-- ;b_flag = 0;
	   }

	   if ( b_flag ) {
	      
	      fill = count-- ;
	      
	      for ( count = fill ; count <= 4 ; count++ ) {
	          
	          fprintf ( Out1 ,"    " );
	      }
	   }
	   else fill = 4 ;
	
	   fprintf (Out1 ,"     ");
	
	   length += fill ;
	
	   for (count = 1;count <=4 ;count++ ) {
	      
	       if (! length) break ;
	      
	       if ( isprint ( *(tmp + Start->packet_length - length)) ) 
		  
		  fprintf ( Out1 ,"%c" ,
		       (unsigned char)*(tmp +Start->packet_length - length) );
		  
	       else fprintf (Out1 ,"*");
			     
	       length--;
			     
	   }
	       
	      
	   fprintf (Out1 ,"\n");   
	      
     }
      
     dummy = Start ;
      
     Start = Start->next ;
      
     free ((void*) dummy->packet_data ) ; 
     
     if (dummy->overwrite) { 
      
        over = dummy->overwrite; 
      
        while (over) { 
      
	       overtmp = over->next ;
               free((void*)over->packet_data);
	       over = overtmp;
	}
     }
      
     free ((void*) dummy ) ;
      
     fflush (stdout); 
     fflush (Out1);

   
  }

  READWRITE = 0 ;   
     
  pthread_cond_signal (&reading) ;  
    
  pthread_mutex_unlock (&mutex);   
    
 }

}


 

void term (int sig) 
{
   
   fflush (NULL) ;
   
   fflush (Out1) ;
   
   fclose (Out1) ;

   close (Fd.fd);
   
   pthread_kill ( thread1ID , 9 ) ;
   
   pthread_kill ( thread2ID , 9 ) ;  
   
   exit(0);
   
}


int main (int argc ,char *argv[]) {

 int count = 0 ,count1 = 0 ,count2 ,number = 0 ;
 unsigned int maxport = 0 ;  
 char *letter ;
 char *iface ;
 char *stop ,*cvn ,*p ;
 struct FLAGS   *Fdummy ;
 struct OPTIONS *Odummy ;
 struct PORT    *Pdummy ;
 int flag = 0 ,ipv_flag = 0 ,offset = 0;

   
 time_t curtime ;
 struct tm *loctime ;
 tzset ();     
 curtime = time (NULL);     
 loctime = localtime (&curtime);    
 Directory = stcat ( "./Pkdump-" , asctime ( loctime ) ,0 ) ;
 p = Directory ;

   
 while (*p) { if ( *p == ' ') *p = '-' ;p++;}
	
 if ( argc < 3 ) {
printf ("\n\n -PKDUMP Ver. 3.3 \n"); 
printf ("Usage : pkdump interface [-di] IP-version [options] flags \n \n"); 
printf ("       interface = ppp0 ,eth0 ,eth1.....etc etc \n");
printf ("      -di = Don't ignore my packet  \n");              
printf ("       IP-version = IPV4 ,IPV6 \n ");
printf ("      Options = [-iT] [startport[,port][:endport]-address/mask] \n ");
printf ("      Options = [-oT] [startport[,port][:endport]-address/mask] \n ");
printf ("      Options = [-aT] [startport[,port][:endport]-address/mask] \n ");
printf ("      Options = [-iU] [startport[,port][:endport]-address/mask] \n ");
printf ("      Options = [-oU] [startport[,port][:endport]-address/mask] \n ");
printf ("      Options = [-aU] [startport[,port][:endport]-address/mask] \n ");
printf ("      IPV4 max mask = 32 ,IPV6 max mask = 128 \n ");
printf ("      Flags = NULL ,UDPNULL,UDPPORT,FIN,SYN,RST,\n"); 
printf ("               PSH,ACK,URG,XMAS,MYMO\n");
printf ("       Flags = flag[:flag]\n\n"); 
    
    printf ("CTRL C (^C) to exit \n\n");
    
    exit (1) ;
 }  
   
 if ( argc >2 ) {  

  *argv++ ;
  IFACE = iface = *argv;  
  *argv++ ;

  if ( *argv) {
    
    if ( !strcmp ( *argv , "-di" ) ) {IGNORE_FLAG = 0 ;*argv++;}
     
  }      
        
  count = 1 ;  
    
  while ( IPversion_name[count] ){  
    
   if ( !strcmp ( *argv ,IPversion_name[count] ) ) { 
	   
	       Ipv.ver = IPversion_value[count] ;
	 
               if ( (Ipv.ver == OIPV4) || (Ipv.ver == OIPV6) ) {
		  
		  *argv++;
      
                  if ( ! *argv) {  
     
                     fprintf (stdout ,"\n\nERROR : Missing flags \n\n");
	             exit (1);
                  }

		  ipv_flag = 1 ;
		  
		  break;
	       }
      
   }
  
   count++ ;  
  
  }
  
  if ( ! ipv_flag ) {  
     
     fprintf (stdout ,"\n\nERROR : Missing IP version \n\n");
     exit (1);
  }  
      
  count = 1 ; 

  while ( Options_name[count] ){  
    
   if ( !strcmp ( *argv ,Options_name[count] ) ) { 
  
	 if ( ! Options ) {
    
                if((Options=Odummy=(struct OPTIONS*) 
		                   malloc(sizeof(struct OPTIONS)))==NULL){
	 fprintf (stdout ,"\n\nERROR :No space for allocating OPTIONS\n\n");
	          exit(1);
                }  

                Options->next = NULL ;
               }
               else {
		     if((Odummy->next=(struct OPTIONS*) 
			            malloc(sizeof(struct OPTIONS)))==NULL){
	 fprintf (stdout ,"\n\nERROR :No space for allocating OPTIONS\n\n");
	              exit(1);
		     }

                     Odummy = Odummy->next ;
	
                     Odummy->next = NULL ;	
               }
	   
	       Odummy->opt = Options_value[count] ;
	 
               *argv++;
     
               cvn = stop = *argv ;  
               for (count1 = 1 ;;count1++) {
		  
		  if ( *stop == '-' ) break ;
		  if ( *stop == ':' ) { flag = 1 ;break ;}
		  if ( *stop == ',' ) { flag = 2 ;break ;}
		  else if (count1 > 6) break ;
		  stop++;
	       }

               if ( count1 > 6 ) {
		  
		  fprintf ( stdout ,"\n\nERROR :Initial Port error \n\n");
		  exit(1);
	       }

               *stop = '\0' ;  
       
               if ( flag == 2 ) {
		  
		  if((Pdummy=Odummy->rport =(struct PORT*) 
		                   malloc(sizeof(struct PORT)))==NULL){
	 fprintf (stdout ,"\n\nERROR :No space for allocating PORTS\n\n");
	          exit(1);
                  }
		  
		  count2 = 1 ;
		  
		  Odummy->rport->port = atoi (cvn) ;
		  Odummy->rport->next = NULL ;

                  if ( Odummy->rport->port > 65535  ) {
		  
		      fprintf ( stdout ,
			   "\n\nERROR :Port %d out of range\n\n",count2);
		      exit(1);
	          }
		  
		  cvn = ++stop ;
		  
		  Odummy->sport=Odummy->eport= 0 ;
		  
		  while (*cvn) {
		     
		        for (count1 = 1 ;;count1++) {
		  
		            if ( *stop == '-' ) { flag = 1 ;break ;}
		            if ( *stop == ',' ) break ;
		            stop++;
	                }
		        count2++;
		     
                        *stop = '\0' ;  
               
		        if ( *(stop + 1) == '-' ) {
		     
		           fprintf ( stdout ,
			     "\n\nERROR : Missing %d port\n\n" ,count2+1);
		           exit(1);
		        }
		     
		        if((Pdummy->next=(struct PORT*) 
		                         malloc(sizeof(struct PORT)))==NULL){
	     fprintf (stdout ,"\n\nERROR :No space for allocating PORTS\n\n");
	                  exit(1);
                        }
		  
		        Pdummy = Pdummy->next ;
		        Pdummy->port = atoi (cvn) ;
		        Pdummy->next = NULL ;

                        if ( Pdummy->port > 65535  ) {
		  
		           fprintf ( stdout ,
			         "\n\nERROR :Port %d out of range\n\n",count2);
		           exit(1);
	                }
     
		        cvn = ++stop ;
		     
		        if ( flag == 1 ) { flag = 0 ;break;}
		       
		  }
		  if ( flag ) {
		     
		     fprintf ( stdout ,
			     "\n\nERROR :Port %d out of range\n\n" ,count2);
		     exit(1);
		  }
		  
		       
	       }/* end if flag==2 */
               else {
		  
		Odummy->rport = NULL ;  
      
                maxport = atoi (cvn);
		  
	        if ( maxport > 65535  ) {
		  
		   fprintf ( stdout ,"\n\nERROR :StartPort out of range\n\n");
		   exit(1);
	        }
      
                Odummy->sport = atoi ( cvn );
       
                cvn = ++stop ;
       
	        if ( flag == 1 ) {

		  flag = 0 ;
		  
		  for (count1 = 1 ;;count1++) {
		  
		      if ( *stop == '-' ) break ;
		      else if (count1 > 6) break ;
		      stop++;
	          }
	          if ( count1 > 6 ) {
		  
		   fprintf ( stdout ,"\n\nERROR :EndPort out of range\n\n");
		   exit(1);
	          }
	 
                  *stop = '\0' ;
	 
		  maxport = atoi (cvn);
		  
		  if ( maxport > 65535  ) {
		  
		   fprintf ( stdout ,"\n\nERROR :EndPort out of range\n\n");
		   exit(1);
	          }
		  
	          Odummy->eport = atoi ( cvn );
		  
		  if ( Odummy->sport >= Odummy->eport ) {
		   fprintf ( stdout ,"\n\nERROR :Wrong Startport Endport\n\n");
		   exit(1);
	          }
		  
	          cvn = ++stop ;
	        }
                else Odummy->eport = 0 ;
	       }/* end else if flag==2 */
      
	       for (count1 = 1 ;;count1++) {
		  
		  if ( *stop == '/') break ;
		  else if (! *stop) break ;
		  stop++;
	       }
               
               if ( *stop != '/' ) {
		  
		  fprintf (stdout ,"\n\nERROR :Missing mask \n\n");
		  exit(1);
	       }

      
               if ( ! *stop  ) {
		  
		  fprintf (stdout ,"\n\nERROR :Address out of range\n\n");
		  exit(1);
	       }
  
	       *stop = '\0' ;

               if ( Ipv.ver == OIPV4 ) {
	   
	          if ( !inet_pton (AF_INET ,cvn,(void*)&(Odummy->myaddress))){
		       
	                fprintf (stdout ,"\n\nERROR :Isn't IPV4 Address \n");
		        exit(1);
		  }
		  fprintf (stdout ,"\nIs IPV4 Address \n");
		  Odummy->myaddress = htonl(Odummy->myaddress) ;
	       }
               else if ( Ipv.ver == OIPV6 ) {
	      
	     if (!inet_pton (AF_INET6 ,cvn ,(void*)&(Odummy->myaddress6[0]))){	     
		     
		     fprintf (stdout ,"ERROR :Isn't IPV6 Address \n\n");
		     exit(1);
	             
	            }
		  
	            fprintf (stdout ,"\nIs IPV6 Address \n");
                    Odummy->myaddress6[0] = htons ( Odummy->myaddress6[0]) ;
                    Odummy->myaddress6[1] = htons ( Odummy->myaddress6[1]) ;
                    Odummy->myaddress6[2] = htons ( Odummy->myaddress6[2]) ;
                    Odummy->myaddress6[3] = htons ( Odummy->myaddress6[3]) ;
                    Odummy->myaddress6[4] = htons ( Odummy->myaddress6[4]) ;
                    Odummy->myaddress6[5] = htons ( Odummy->myaddress6[5]) ;
                    Odummy->myaddress6[6] = htons ( Odummy->myaddress6[6]) ;
                    Odummy->myaddress6[7] = htons ( Odummy->myaddress6[7]) ;
		   
	       }
      
               cvn = ++stop ;
	
	       for (count1 = 1 ;;count1++) {
		  
		  if ( *stop == '\0') break ;
		  else if ( count1 > 3 && Ipv.ver == OIPV4 ) break ;
		       else if ( count1 > 4 && Ipv.ver == OIPV6 ) break ;
		  stop++ ;
	       }
	       if ( count1 == 1 ) {
		  
		  fprintf (stdout ,"\n\nERROR :Missing mask\n\n");
		  exit(1);
	       }
      
               if ( count1 > 3 && Ipv.ver == OIPV4  ) {
		  
		  fprintf (stdout ,"\n\nERROR :IPV4 Mask out of range\n\n");
		  exit(1);
	       }
               else if ( count1 > 4 ) {
		  
		  fprintf (stdout ,"\n\nERROR :IPV6 Mask out of range\n\n");
		  exit(1);
	       }
	       
	       *stop = '\0' ;
	
	       Odummy->mask = atoi ( cvn );
	
	       if ( Odummy->mask > 32 && Ipv.ver == OIPV4 ) {
		  
		  fprintf (stdout ,"\n\nERROR :IPV4 Mask out of range\n\n");
		  exit(1);
	       }
               else if ( Odummy->mask > 128 ) {
		  
		  fprintf (stdout ,"\n\nERROR :IPV6 Mask out of range\n\n");
		  exit(1);
	       }

               *argv++;
      
               if ( ! *argv) {  
     
                  fprintf (stdout ,"\n\nERROR : Missing flags \n\n");
	          exit (1);
               }

      
	       count = 1 ;continue ;
	 
      
   }
     
   count++ ;  
     
  }   

    
  while ( *argv ) { 
	 
    letter = *argv ;
     
    while ( *letter ) { 
	    if (*letter >= 0x61) *letter -= 0x20 ;
	    letter ++;
    }
     
    offset = flag = count = 0 ; 

    if (!Flags) {	 
         
       if((Flags=Fdummy=(struct FLAGS*)malloc(sizeof(struct FLAGS)))==NULL){
	  fprintf (stdout ,"\n\nERROR :No space for allocating FLAGS\n\n");
	  exit(1);
       }
       Fdummy->flags = 0 ;  
       Fdummy->next  = NULL ;
         
    }	 
    else {
	   
	 if ((Fdummy->next=(struct FLAGS*)malloc(sizeof(struct FLAGS)))==NULL){
	    fprintf (stdout , "\n\nERROR :No space for allocating FLAGS\n\n");
	    exit(1);
	 }    
	  
	 Fdummy = Fdummy->next ;
	 Fdummy->flags = 0 ; 
	 Fdummy->next  = NULL ;
	     
    }  	 
 
    do { 	  
       
     if (!strncmp ( &(**argv)+offset ,"\0"  ,1 ) ) break ;  
       
     while ( Flags_name[count] ) {	 
     
      if (!strncmp ( &(**argv)+offset ,Flags_name[count]  ,3 ) ) {	 
       	 
       flag = 1 ;	  
	  	       
       Fdummy->flags |= Flags_value[count] ;
	 
       break ;	 
      }
	
      count ++;
     
     }  
     
     if ( ! flag  ) {
	  fprintf (stdout ,"\n\nERROR : Flags don't recognized \n\n");
	  exit (1);
     }
     else flag = 0 ;
       
     if ( !strncmp ( &(**argv)+offset+3 , ":" ,1 ) ) {count = 0 ;offset += 4 ;}
     else break ;	
       
    }while (1);
      	       
    *argv++;
  } 
    
 }  
     
       
 Fd.fd = if_pr_set ( iface ) ;
   
 pthread_setconcurrency (2) ;
 
   
   
 pthread_create (&thread1ID ,NULL ,thread_read  ,NULL ) ;  

 pthread_create (&thread2ID ,NULL ,thread_write ,NULL ) ;  
   
 
 signal ( SIGINT ,term );  
   
   
 pthread_join ( thread2ID ,NULL ) ;

   
}

   
    
